from __future__ import annotations
import warnings
from typing import Optional, Union, Callable, List

import numpy as np
import matplotlib
from matplotlib import pyplot as plt
import shutil

from michael960lib.common import IllegalActionError, scalarize
from torusgrid import fields as fd
import torusgrid as tg
from .core.evolution import ConstantChemicalPotentialMinimizer, NonlocalConservedMinimizer, StressRelaxer, PFCMinimizer
from .core.base import PFCStateFunction
from .history import PFCHistory, PFCMinimizerHistoryBlock, PFCEditActionHistoryBlock, import_history

matplotlib.use('TKAgg')
matplotlib.style.use('fast')


class PFC:
    def __init__(self, field: fd.RealField2D):

        self.field = field 
        self.age = 0
        self.history= PFCHistory(self.field)

        self.history_pointer = 0
        self.current_minimizer = None

    def new_minimizer(self, minimizer: PFCMinimizer):
        try:
            assert isinstance(minimizer, PFCMinimizer)
            assert not minimizer.started
            assert not minimizer.ended
            assert not minimizer.history.is_committed()
            assert minimizer.age == 0
            assert minimizer.field is self.field
        except AssertionError:
            raise ValueError('invalid minimizer')

        self.current_minimizer = minimizer

    def new_mu_minimizer(self, dt, eps, mu):
        self.current_minimizer = ConstantChemicalPotentialMinimizer(self.field, dt, eps, mu)

    def new_nonlocal_minimizer(self, dt, eps):
        self.current_minimizer = NonlocalConservedMinimizer(self.field, dt, eps)

    def new_stress_relaxer(self, dt, eps, mu, expansion_rate=1):
        self.current_minimizer = StressRelaxer(self.field, dt, eps, mu)

    def evolve_multisteps(self, N_steps, N_epochs,
            display_format: Optional[str]=None,
            callbacks: List[Callable[[PFCMinimizer, PFCStateFunction], None]]=[]):

        if self.current_minimizer is None:
            raise tg.dynamics.MinimizerError(self.current_minimizer) 

        self.current_minimizer.set_display_format(display_format)
        self.current_minimizer.run_multisteps(N_steps, N_epochs, callbacks=callbacks)

        self.history_pointer += 1
        self.age += self.current_minimizer.age
        self.history.cut_and_insert(PFCMinimizerHistoryBlock(self.current_minimizer.history), self.history_pointer)
        self.current_minimizer = None

    def evolve_nonstop(self, N_steps, custom_keyboard_interrupt_handler=None,
            display_format: Optional[str]=None,
            callbacks: List[Callable[[PFCMinimizer, PFCStateFunction], None]]=[]):

        if self.current_minimizer is None:
            raise tg.dynamics.MinimizerError(self.current_minimizer) 

        self.current_minimizer.set_display_format(display_format)
        self.current_minimizer.run_nonstop(N_steps, custom_keyboard_interrupt_handler=custom_keyboard_interrupt_handler, 
                callbacks=callbacks)

        self.history_pointer += 1
        self.age += self.current_minimizer.age
        self.history.cut_and_insert(PFCMinimizerHistoryBlock(self.current_minimizer.history), self.history_pointer)
        self.current_minimizer = None

    def evolve(self,
        minimizer_supplier: Optional[Callable[[PFC], PFCMinimizer]]=None,
        minimizer: str=None, dt: float=None, eps: float=None, mu: Optional[float]=None,
        expansion_rate: Optional[float]=None,
        N_steps: int=31, N_epochs: Optional[int]=None,
        custom_keyboard_interrupt_handler: Optional[Callable[[PFCMinimizer], bool]]=None,
        display_format: Optional[str]=None, 
        callbacks: List[Callable[[PFCMinimizer, PFCStateFunction], None]]=[]):

        if not minimizer_supplier is None:
            if None not in (minimizer, dt, eps, mu, expansion_rate):
                warnings.warn('ignoring other arguments passed to evolve() when using minimizer supplier')

            minim = minimizer_supplier(self)
            self.new_minimizer(minim)

        else:
            if not minimizer in ['mu', 'nonlocal', 'relax']:
                raise ValueError(f'{minimizer} is not a valid minimizer')

            if N_steps <= 0:
                raise ValueError(f'N_steps must be a positive integer')
            
            if minimizer == 'mu':
                if mu is None:
                    raise ValueError(f'chemical potential must be specified with constant chemical potential minimizer')

                if not (expansion_rate is None):
                    warnings.warn(f'expansion rate will be ignored for constant chemical potential minimizer')

                self.new_mu_minimizer(dt, eps, mu)
            if minimizer == 'nonlocal':
                if not (mu is None):
                    warnings.warn(f'chemical potential will be ignored for nonlocal conserved minimizer')
                if not (expansion_rate is None):
                    warnings.warn(f'expansion rate will be ignored for nonlocal conserved minimizer')

                self.new_nonlocal_minimizer(dt, eps)

            if minimizer == 'relax':
                if mu is None:
                    raise ValueError(f'chemical potential must be specified with constant mu stress relaxer')
                if expansion_rate is None:
                    raise ValueError(f'expansion rate must be specified with constant mu stress relaxer')

                self.new_stress_relaxer(dt, eps, mu, expansion_rate=expansion_rate)


        if N_epochs is None:
            self.evolve_nonstop(N_steps, custom_keyboard_interrupt_handler=custom_keyboard_interrupt_handler,
                    display_format=display_format, callbacks=callbacks)
        else:
            if N_epochs <=0:
                raise ValueError(f'N_epochs must be a positive integer')
            self.evolve_multisteps(N_steps, N_epochs, display_format=display_format,
                    callbacks=callbacks)

    def field_snapshot(self):
        return self.field.export_state()     

    def plot_history(self, *item_names, show=True):
        if len(item_names) == 0:
            item_names = ['f', 'psibar']
        return self.history.plot(*item_names, start=0, end=self.history_pointer, show=show)
    
    def undo(self):
        if self.history_pointer <= 0:
            self.history_pointer = 0
            raise IllegalActionError('history is already at the oldest state')

        self.history_pointer -= 1
        field_state = self.history.get_block(self.history_pointer).get_final_field_state()
        self.field.set_psi(field_state['psi'])
        self.field.set_size(field_state['Lx'], field_state['Ly'])
    
    def redo(self):
        if self.history_pointer >= len(self.history.blocks)-1:
            self.history_pointer = len(self.history.blocks)-1
            raise IllegalActionError('history is already at the newest state')

        self.history_pointer += 1
        field_state = self.history.get_block(self.history_pointer).get_final_field_state()
        self.field.set_psi(field_state['psi'])
        self.field.set_size(field_state['Lx'], field_state['Ly'])
    
    def export(self) -> dict:
        state = dict()
        state['history'] = self.history.export()
        state['age'] = self.age
        state['history_pointer'] = self.history_pointer
        state['field'] = self.field.export_state()

        return state
    
    def save(self, path: str):
        state = self.export()

        tmp_name = f'{path}.tmp.file'
        np.savez(tmp_name, state=state)
        shutil.move(f'{tmp_name}.npz', f'{path}.pfc')

    def save_hdf5(self, path):
        raise NotImplementedError
        state = self.export()


def import_pfc_model(state: dict) -> PFC:
    history_state = state['history']
    field_state = state['field']

    pfc_history = import_history(history_state) 
    field = fd.import_field(field_state)

    pfc_model = PFC(field)

    pfc_model.age = state['age']
    pfc_model.history_pointer = state['history_pointer']
    pfc_model.history = pfc_history 

    return pfc_model


def load_pfc_model(path: str) -> PFC:
    data = np.load(path, allow_pickle=True)
    return import_pfc_model(scalarize(data['state']))


class PFCGroup:
    def __init__(self):
        self.models = dict()
    
    def put(self, model: PFC, name: str, attrs=dict()):
        if name in self.get_names():
            warnings.warn(f'key {name} already exists in group and will be overwritten')
        self.models[name] = {'model': model, 'attrs': attrs}

    def save(self, path):
        data = dict() 
        tmp_name = f'{path}.tmp.file'
        for name in self.models:
            model_state = self.models[name]['model'].export()
            attrs = self.models[name]['attrs']
            data[name] = {'model': model_state, 'attrs': attrs}
        np.savez(tmp_name, **data)
        shutil.move(f'{tmp_name}.npz', f'{path}.pfcgroup')

    def get_names(self) -> List[str]:
        return list(self.models.keys())

    def get_model(self, name) -> PFC:
        return self.models[name]['model']

    def get_attrs(self, name) -> dict:
        return self.models[name]['attrs']

    def get_models(self) -> list:
        l = []
        for key in self.models:
            l.append(self.models[key]['model'])
        return l

    def get_sorted_models(self, comparator: str, reverse: bool=False) -> List[PFC]:
        def get_item(model):

            try:
                val = {
                        'Lx': model.field.Lx,
                        'Ly': model.field.Ly,
                        'Nx': model.field.Nx,
                        'Ny': model.field.Ny
                }[comparator]
            except KeyError:
                raise ValueError(f'{comparator} is not a valid attribute')
            return val

        models = self.get_models()
        return sorted(models, key=get_item, reverse=reverse)


def load_pfc_group(path: str) -> PFCGroup:
    saved = np.load(path, allow_pickle=True)
    g = PFCGroup()

    for name in saved.files:
        data = scalarize(saved[name])
        model_state = data['model']
        model = import_pfc_model(model_state)
        attrs = data['attrs']

        g.put(model, name, attrs=attrs)
    return g




#!/opt/anaconda3/envs/turnbull/bin/python3
from ..legacy import pfc_im
from matplotlib import pyplot as plt
import numpy as np
import sys
import os
import subprocess
import readline
from util import common as cmn
import os
import re
import readline


_version = '2.2'
RE_SPACE = re.compile('.*\s+$', re.M)



class SimplePrompt(object):
    def __init__(self):
        self.commands_obj = dict()
        self.commands_str = ''
        self.running = False
        self.prefix = '>'

        readline.parse_and_bind('set show-all-if-ambiguous on')
        readline.parse_and_bind('set colored-stats on')
        readline.parse_and_bind('tab: menu-complete')


        readline.set_completer_delims(' \t\n;')
        readline.set_completer(self.complete)



    def add_command(self, command):
        self.commands_obj[command.name] = command
        self.commands_str = ', '.join((f'{cmd}' for cmd in self.commands_obj.keys()))

    def consume(self):
        cmd_str = input(f'{self.prefix }')
        self.handle_command(cmd_str)

    def handle_command(self, cmd_str: str):
        cmd_and_args = cmd_str.split()
        
        if len(cmd_and_args) == 0:
            return

        cmd = cmd_and_args[0]
        args = cmd_and_args[1:]

        if cmd not in self.commands_obj:
            self.output(f'command {cmd} not found, possible commands are', level=-1)
            self.output(self.commands_str, level=-1)
            print()
            return
        
        self.execute_command(cmd, args)

    def execute_command(self, cmd, args):
        try:
            cmd = self.commands_obj[cmd]
            cmd.execute(self, args)
            return

        except CommandExecutionError as e:
            self.output(e.message, 2)
            print()

            
    def start(self):
        self.put_banner()
        self.running = True
        while self.running:
            try:
                self.consume()
            except KeyboardInterrupt:
                print()
                self.output('enter exit to leave', level=1)
                print()
                pass
        self.output('terminating prompt')
        self._on_exit()

    def put_banner(self):
        pass

    # 0: ok, 1: warning, else: error
    def output(self, s, level=0):
        if level == 0:
            print(f'{cmn.bcolors.BOLD}{cmn.bcolors.OKCYAN}[]{cmn.bcolors.ENDC} {s}')
        elif level == 1:
            print(f'{cmn.bcolors.BOLD}{cmn.bcolors.WARNING}[]{cmn.bcolors.ENDC} {s}')
        else:
            print(f'{cmn.bcolors.BOLD}{cmn.bcolors.FAIL}[]{cmn.bcolors.ENDC} {s}')

    def complete(self, text, state):
        # Generic readline completion entry point
        buff = readline.get_line_buffer()
        line = readline.get_line_buffer().split()

        # show all commands
        if not line:
            return [c + ' ' for c in self.commands_obj.keys()][state]

        # account for last argument ending in a space
        if RE_SPACE.match(buff):
            line.append('')

        # resolve command to the implementation function
        cmd = line[0].strip()
        if cmd in self.commands_obj.keys():
            args = line[1:]
            if args:
                candidates = self.commands_obj[cmd].complete(self, args)
                result_list = [can for can in candidates if can.startswith(args[-1])] + [None]
                
                return result_list[state]
            return [cmd + ' '][state]


        # incomplete command
        results = [c + ' ' for c in self.commands_obj.keys() if c.startswith(cmd)] + [None]
        return results[state]

    def _on_exit(self):
        return



class CommandExecutionError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(message)


class Command:
    def __init__(self, name):
        self.name = name
        self._execute = lambda args: None
        self._completer = None
        self._prompt = None

    def execute(self, prompt, args):
        self._execute(prompt, args)

    def complete(self, prompt, args):
        return _complete_path(args)



def _listdir(root):
    "List directory 'root' appending the path separator to subdirs."
    res = []
    for name in os.listdir(root):
        path = os.path.join(root, name)
        if os.path.isdir(path):
            name += os.sep
        res.append(name)
    return res

def _complete_path(args):
    path = '.'    
    if not args:
        path = '.'
    else:
        path = args[-1]

    "Perform completion of filesystem path."
    if not path:
        return _listdir('.')

    dirname, rest = os.path.split(path)
    tmp = dirname if dirname else '.'
    res = [os.path.join(dirname, p)
            for p in _listdir(tmp) if p.startswith(rest)]

    # more than one match, or single match which does not exist (typo)
    if len(res) > 1 or not os.path.exists(path):
        return res
    # resolved to a single directory, so return list of files below it
    if os.path.isdir(path):
        return [os.path.join(path, p) for p in _listdir(path)]
    # exact file match terminates this completion
    return [path + ' ']





import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import fft2, ifft2, rfft2, irfft2, set_global_backend
from pprint import pprint
import pyfftw
import time
import threading
import sys

from michael960lib.math import fourier
from michael960lib.common import overrides


# pyfftw implementation
class PhaseFieldCrystal2D:
    def __init__(self, Lx, Ly, Nx, Ny, verbose=True):

        self.set_dimensions(Lx, Ly, Nx, Ny, verbose)
        self.set_psi(np.zeros((Nx, Ny)), verbose=verbose)
        self.set_params(0, 0, verbose=verbose)
        self.set_noise(0, 0)
        self.lock = None
        self.new_lock()
        self.new_history()

        #self.minimizer = 'mu'
        #self.running = False

        set_global_backend(pyfftw.interfaces.scipy_fft)

    def set_noise(self, seed, width):
        self.seed = seed
        np.random.seed(self.seed)
        self.noise_width = width

    def set_params(self, mu, eps, verbose=True):
        self.mu = mu
        self.eps = eps
        if verbose:
            self.yell(f'new params set, mu={mu} eps={eps}')

    def set_dimensions(self, Lx, Ly, Nx, Ny, verbose=True):

        self.Nx = Nx
        self.Ny = Ny
        
        self.resize(Lx, Ly, verbose=verbose)

        self.psi = pyfftw.zeros_aligned((Nx, Ny), dtype='float64')
        self.psi_k = pyfftw.zeros_aligned((Nx, Ny//2+1), dtype='complex128')

        if verbose:
            self.yell(f'new resolution set, Nx={Nx} Ny={Ny}')
            self.yell(f'reset profile, Lx={Lx} Ly={Ly} Nx={Nx} Ny={Ny}')

    def resize(self, Lx, Ly, verbose=True):
        self.Lx = Lx
        self.Ly = Ly
        self.Volume = Lx*Ly
        x, kx, self.dx, self.dkx, y, ky, self.dy, self.dky = fourier.generate_xk_2d(Lx, Ly, self.Nx, self.Ny, real=True)

        self.X, self.Y = np.meshgrid(x, y, indexing='ij')
        self.Kx, self.Ky = np.meshgrid(kx, ky, indexing='ij')
        self.K2 = self.Kx**2 + self.Ky**2
        self.K4 = self.K2**2

        self.Kernel = 1-2*self.K2+self.K4

        if verbose:
            self.yell(f'new dimensions set, Lx={Lx} Ly={Ly}') 
            self.yell(f'k-space dimensions {self.K4.shape}')

    def initialize_fftw(self, **fftwargs):
        psi_tmp = self.psi.copy()
        self.for_psi = pyfftw.FFTW(self.psi, self.psi_k, direction='FFTW_FORWARD', axes=(0,1), **fftwargs)
        self.bac_psi = pyfftw.FFTW(self.psi_k, self.psi, direction='FFTW_BACKWARD', axes=(0,1), **fftwargs)
        self.set_psi(psi_tmp)

    def set_psi(self, psi1, verbose=True):
        if psi1.shape[0] != self.Nx or psi1.shape[1] != self.Ny:
            raise ValueError(f'array has incompatible shape {psi1.shape} with ({self.Nx, self.Ny})')

        self.psi[:,:] = psi1
        if verbose:
            self.yell('new psi set')

    def new_history(self):
        #self.history = {'t': [], 'Omega': [], 'psi0': [], 'max-min': [], 'omega': []}
        self.history = []

    def load_profile_from_file(self, saved_npz, verbose=True):
        file_list = saved_npz.files
        if verbose:
            self.yell(f'loading npz file, found the following data:')
            self.yell(f'{file_list}')

        psi = saved_npz['psi']

        self.set_dimensions(saved_npz['Lx'], saved_npz['Ly'], *psi.shape, verbose=verbose)

        self.set_params(saved_npz['mu'], saved_npz['eps'], verbose=verbose)
        self.set_psi(saved_npz['psi'], verbose=verbose)
        self.dt = saved_npz['dt']

        try:
            if saved_npz['history'] is None:
                raise Exception
            self.history = saved_npz['history'].reshape((1,))[0]
        except:
            self.new_history()
            if verbose:
                self.yell('no history file found, initialized empty history')

        if saved_npz['minimizer'] is None:
            self.minimizer = 'mu'
        else:
            self.minimizer = saved_npz['minimizer']

    def dump_profile_to_file(self, target_npz, verbose=True, **kwargs):
        if verbose:
            self.yell(f'dumping profile data to {target_npz}')

        np.savez(target_npz, psi=self.psi, Lx=self.Lx, Ly=self.Ly, 
                 mu=self.mu, eps=self.eps, history=self.history, 
                 dt=self.dt, minimizer=self.minimizer, **kwargs)

    def _evolve_noise(self):
        self.psi += self.dt * np.random.normal(0, self.noise_width, size=(self.Nx, self.Ny))

    def calc_N_tot(self):
        # total particle number = psi integrated over all volume
        return np.sum(np.real(self.psi)) * self.dx * self.dy

    # local chemical potential
    def calc_chemical_potential_density(self):
        D2psi = irfft2(-self.K2*rfft2(self.psi))
        D4psi = irfft2(self.K4*rfft2(self.psi))
        local_mu = (1-self.eps) * self.psi + self.psi**3 + 2*D2psi + D4psi

        return np.real(local_mu)

    # local grand potential density
    def calc_grand_potential_density(self):
        return self.calc_helmholtz_density() - self.mu * self.psi

    def calc_grand_potential(self):
        return np.sum(self.calc_grand_potential_density()) * self.dx * self.dy
    
    # local helmholtz free energy density
    def calc_helmholtz_density(self):
        psi_k = rfft2(self.psi)
        psi_k_o = self.Kernel * psi_k
        f = 1/2 * self.psi * irfft2(psi_k_o) + self.psi**4/4 - self.eps/2 * self.psi**2
        return np.real(f)

    # difference
    def calc_ord_param(self):
        return np.max(self.psi) - np.min(self.psi)
    
    # plot psi, chemical potential density, and grand potential density
    def plot(self, cmap='jet', plot_psi=True, plot_mu=True, plot_omega=True, lazy_factor=1):

        num_plots = plot_psi + plot_mu + plot_omega

        plt.figure(figsize=(12.8, 3.2*num_plots))
        index_plot = 1

        LF = lazy_factor

        if plot_psi:
            ax1 = plt.subplot(num_plots, 1, index_plot, title='$\\psi$')
            cm1 = ax1.pcolormesh(self.X[::LF], self.Y[::LF], np.real(self.psi)[::LF], cmap=cmap, vmin=-1, vmax=1, shading='nearest')
            plt.colorbar(cm1, ax=ax1, orientation='horizontal')
            ax1.set_aspect('equal', adjustable='box')

            index_plot += 1

        if plot_mu:

            ax2 = plt.subplot(num_plots, 1, index_plot, title='$\\mu(\\mathbf{r})-\\mu$')
            cm2 = ax2.pcolormesh(self.X[::LF], self.Y[::LF], self.calc_chemical_potential_density()[::LF] - self.mu, cmap=cmap, vmin=-0.2, vmax=0.2, shading='nearest')
            plt.colorbar(cm2, ax=ax2, orientation='horizontal')
            ax2.set_aspect('equal', adjustable='box')

            index_plot += 1

        if plot_omega:

            ax3 = plt.subplot(num_plots, 1, index_plot, title='$\\omega(\\mathbf{r})$')
            cm3 = ax3.pcolormesh(self.X[::LF], self.Y[::LF], self.calc_grand_potential_density()[::LF], cmap=cmap, vmin=-0.1, vmax=0.1, shading='nearest')
            plt.colorbar(cm3, ax=ax3, orientation='horizontal')
            ax3.set_aspect('equal', adjustable='box')

            index_plot += 1


        plt.tight_layout()
        plt.show(block=True)

    def _prepare_simulation(self, dt):
        self._exp_dt_kernel = np.exp(-dt*self.Kernel)
        self._exp_dt_eps = np.exp(dt*self.eps/2)
        self.dt = dt
        self.initialize_fftw()

        # used for nonlocal convserved
        self._psi_bar = np.mean(self.psi)
        self._exp_dt_kernel_conserved = self._exp_dt_kernel.copy()
        self._exp_dt_kernel_conserved[0,0] = 1.
        
    def _evolve_k(self):
        self.for_psi()
        self.psi_k *= self._exp_dt_kernel
        self.bac_psi()

    def _evolve_x(self):
        self.psi *= self._exp_dt_eps
        self.psi -= self.dt/2 * (self.psi**3 - self.mu)

    def _evolve_conserved_nonlocal(self):
        dt = self.dt

        #psi3 = self.psi**3
        #self.psi -= dt/2 * (psi3 - np.mean(psi3))
        #self.psi = (self.psi - self._psi_bar)*self._exp_dt_eps + self._psi_bar

        #self.for_psi()
        #self.psi_k *= self._exp_dt_kernel_conserved
        #self.bac_psi()

        #psi3 = self.psi**3
        #self.psi -= dt/2 * (psi3 - np.mean(psi3))
        #self.psi = (self.psi - self._psi_bar)*self._exp_dt_eps + self._psi_bar
        
        psi3k = rfft2(self.psi**3)

        self.for_psi()
        psik00 = self.psi_k[0,0]
        self.psi_k -= dt * psi3k
        self.psi_k *= self._exp_dt_eps**2
        self.psi_k *= self._exp_dt_kernel_conserved
        self.psi_k[0,0] = psik00
        self.bac_psi()
        self.psi = np.real(self.psi)

    #
    def summarize(self, display_precision=5):
        dp = display_precision
        self.yell(f'------------------------------------------------------------')
        self.yell(f'Summary:')
        self.yell(f'mu={self.mu}\t\teps={self.eps}')
        self.yell(f'Lx={self.Lx:.{dp}f}\t\tLy={self.Ly:.{dp}f}')
        self.yell(f'Nx={self.Nx}\t\tNy={self.Ny}')
        dpu = self.Nx * self.Ny / self.Volume * np.pi*4 * np.pi*4/np.sqrt(3)
        self.yell(f'DPUC={dpu:.{dp}f}')
        self.yell(f'minimizer={self.minimizer}\tdt={self.dt}')

        if len(self.history['t']) > 0:
            age = self.history['t'][-1]
            self.yell(f'age={age}')
        else:
            self.yell(f'age=not evolved yet')

        Omega = self.calc_grand_potential()
        omega = Omega / self.Volume
        psi0 = self.calc_N_tot() / self.Volume
        self.yell(f'Omega={Omega:.{dp}f}\tomega={omega:.{dp}f}\tpsi0={psi0:.{dp}f}')
        self.yell(f'------------------------------------------------------------')

    def new_lock(self):
        self.lock = threading.Lock()
        return self.lock

    def destroy_lock(self):
        self.lock = None

    def stop_minimization(self):
        self.running = False


    # minimize with constant mu
    def minimize_mu(self, dt, cycle=31, verbose=True, display_precision=5):
        i = 0
        t = 0
        if len(self.history['t']) > 0:
            t = self.history['t'][-1]

        self._prepare_simulation(dt)
        self.running = True

        dp = display_precision

        if verbose:
            self.summarize()
        
        while True:
            with self.lock:
                if not self.running:
                    return
                self._evolve_noise()
                self._evolve_x()
                self._evolve_k()
                self._evolve_x()
            t += dt
            i += 1

                                
            if i >= cycle:
                psi0 = self.calc_N_tot() / self.Volume
                omega = self.calc_grand_potential_density()
                Omega = np.sum(omega) * self.dx * self.dy
                ord_param = self.calc_ord_param()
                mu = np.mean(self.calc_chemical_potential_density())
                with self.lock:
                    if not self.running:
                        return
                    sys.stdout.write(
                        f'\r[pfcim] t={t:.{dp}f} | psi_bar={psi0:.{dp}f} | Omega={Omega:.{dp}f} | ' +\
                        f'omega={Omega/self.Volume:.{dp+2}f} | mu={mu:.{dp+2}f} | diff={ord_param:.{dp}f}       '
                    )

                self.history['Omega'] = np.append(self.history['Omega'], Omega)
                self.history['t'] = np.append(self.history['t'], t)
                self.history['psi0'] = np.append(self.history['psi0'], psi0)
                self.history['max-min'] = np.append(self.history['max-min'], ord_param)
                self.history['omega'] = np.append(self.history['omega'], Omega/self.Volume)
                i = 0


        print()

    # minimize with nonlocal conserved dynamics
    def minimize_nonlocal_conserved(self, dt, cycle=31, verbose=True, display_precision=5):
        i = 0
        t = 0
        if len(self.history['t']) > 0:
            t = self.history['t'][-1]

        self._prepare_simulation(dt)
        self.running = True

        dp = display_precision

        if verbose:
            self.summarize()
        
        while True:
            with self.lock:
                if not self.running:
                    return
                self._evolve_noise()
                self._evolve_conserved_nonlocal()
            t += dt
            i += 1
                                
            if i >= cycle:
                psi0 = self.calc_N_tot() / self.Volume
                omega = self.calc_grand_potential_density()
                Omega = np.sum(omega) * self.dx * self.dy
                helm = self.calc_helmholtz_density()
                Helm = np.sum(helm) * self.dx * self.dy
                ord_param = self.calc_ord_param()
                mu = np.mean(self.calc_chemical_potential_density())
                with self.lock:
                    if not self.running:
                        return
                    sys.stdout.write(
                        f'\r[pfcim] t={t:.{dp}f} | psi_bar={psi0:.{dp}f} | Omega={Omega:.{dp}f} | ' +\
                        f'omega={Omega/self.Volume:.{dp+2}f} | mu={mu:.{dp+2}f} | diff={ord_param:.{dp}f}       '
                    )

                self.history['Omega'] = np.append(self.history['Omega'], Omega)
                self.history['t'] = np.append(self.history['t'], t)
                self.history['psi0'] = np.append(self.history['psi0'], psi0)
                self.history['max-min'] = np.append(self.history['max-min'], ord_param)
                self.history['omega'] = np.append(self.history['omega'], Omega/self.Volume)
                i = 0

        print()

    def run_background(self, func, args, kwargs=dict()):
        if self.lock is None:
            self.new_lock()
        
        thread = threading.Thread(target=func, args=args, kwargs=kwargs, daemon=False)
        thread.start()

        return self.lock, thread
    
    def copy(self, verbose=True, copy_history=True):
        other = PhaseFieldCrystal2D(1, 1, 1, 1, verbose=verbose)
        other.set_dimensions(self.Lx, self.Ly, self.Nx, self.Ny, verbose=verbose)
        other.set_params(self.mu, self.eps, verbose=verbose)
        other.set_psi(self.psi, verbose=verbose)

        other.dt = self.dt
        if copy_history:
            other.history = self.history.copy()
        else:
            other.new_history()

        return other

    def yell(self, s):
        print(f'[pfcim] {s}')





from numbers import Number
from typing import Tuple

import numpy as np
from torusgrid.fields import import_field, RealField2D

from . import _res

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources



_uc_path_map = {
    '0.04': 'uc_eps0.04.field',
    '0.06': 'uc_eps0.06.field',
    '0.08': 'uc_eps0.08.field',
    '0.1': 'uc_eps0.1.field',
    '0.2': 'uc_eps0.2.field',
    '0.3': 'uc_eps0.3.field',
    '0.4': 'uc_eps0.4.field'
}

_liq_path_map = {
    '0.04': 'uc_eps0.04_liq.field',
    '0.06': 'uc_eps0.06_liq.field',
    '0.08': 'uc_eps0.08_liq.field',
    '0.1': 'uc_eps0.1_liq.field',
    '0.2': 'uc_eps0.2_liq.field',
    '0.3': 'uc_eps0.3_liq.field',
    '0.4': 'uc_eps0.4_liq.field'
}

def get_relaxed_minimized_coexistent_unit_cell(eps: str, liquid=False) -> RealField2D:
    if not type(eps) is str:
        raise ValueError(f'eps should be specified as a string, not {type(eps)}')

    if liquid:
        if not eps in _liq_path_map.keys():
            raise ValueError(f'no static resource for eps=\'{eps}\' liquid currently')

        with pkg_resources.path(_res, _liq_path_map[eps]) as pth:
            field = import_field(np.load(pth, allow_pickle=True))
        return field

    else:
        if not eps in _uc_path_map.keys():
            raise ValueError(f'no static resource for eps=\'{eps}\' solid currently')

        with pkg_resources.path(_res, _uc_path_map[eps]) as pth:
            field = import_field(np.load(pth, allow_pickle=True))
        return field


    raise ValueError(f'no static resource for eps=\'{eps}\' currently')


_coex_epsmu_bounds = {
        '0.04': (0.125627867, 0.125627868),
        '0.06': (0.152877453, 0.152877454),
        '0.08': (0.175436493, 0.175436494),
        '0.1': (0.194970474, 0.194970476), 
        '0.2': (0.268182615, 0.268182617),
        '0.3': (0.320131817, 0.320131819),
        '0.4': (0.360511484, 0.360511486)
}

_coex_epsmu_final = {
    '0.04': 0.12562786787748337,
    '0.06': 0.152877453515625,
    '0.08': 0.1754364934563637,
    '0.1': 0.19497047539062498,
    '0.2': 0.268182616015625,
    '0.3': 0.320131818359375,
    '0.4': 0.360511484765625
}

_coex_epsucf = {
        '0.1': 1.00047017737,
        '0.2': 1.00174100766,
        '0.3': 1.00341955414,
        '0.4': 1.00521892348
}

def get_coexistent_mu_bounds(eps: str) -> Tuple[float, float]:
    if not eps in _coex_epsmu_bounds.keys():
        return ValueError(f'no coexistent mu for eps=\'{eps}\' currently')
    return _coex_epsmu_bounds[eps]

def get_coexistent_mu_final(eps: str) -> float:
    if not eps in _coex_epsmu_final.keys():
        return ValueError(f'no coexistent mu for eps=\'{eps}\' currently')
    return _coex_epsmu_final[eps]

def get_relaxed_unit_cell_size(eps: str, ratio=True) -> Tuple[float, float]:
    if not type(eps) is str:
        raise ValueError(f'eps should be specified as a string, not {type(eps)}')

    if not eps in _uc_path_map.keys():
        raise ValueError(f'no static resource for eps=\'{eps}\' liquid currently')

    field = get_relaxed_minimized_coexistent_unit_cell(eps)
    if ratio:
        return (field.Lx / np.pi / 4, field.Ly / np.pi / 4 * np.sqrt(3))
    else:
        return (field.Lx, field.Ly)









import numpy as np
from scipy.fft import rfft2, irfft2

from michael960lib.common import overrides, deprecated, experimental
from torusgrid.fields import RealField2D
from torusgrid.dynamics import FreeEnergyFunctional2D, NoiseGenerator2D
from ..core.base import PFCFreeEnergyFunctional
from ..core.evolution import ConstantChemicalPotentialMinimizer


class GaussianPFCFucntional(PFCFreeEnergyFunctional):
    def __init__(self, eps):
        self.eps = eps

    @overrides(FreeEnergyFunctional2D)
    def free_energy_density(self, field: RealField2D):
        kernel = np.exp(1-field.K2) + field.K2 + 1 - np.e
        psi_k = rfft2(field.psi)
        psi_k_o = kernel * psi_k
        f = 1/2 * field.psi * irfft2(psi_k_o) + field.psi**4/4 - self.eps/2 * field.psi**2
        return np.real(f)

    @overrides(FreeEnergyFunctional2D)
    def derivative(self, field: RealField2D):
        kernel = np.exp(1-field.K2) + field.K2 + 1 - np.e
        field.fft2()
        linear_term = (kernel - self.eps) * field.psi_k
        field.ifft2()
        local_mu = irfft2(linear_term) + field.psi**3
        return local_mu



class GaussianPFCMinimizer(ConstantChemicalPotentialMinimizer):
    def __init__(self, field: RealField2D, dt: float, eps: float, mu: float, noise_generator:NoiseGenerator2D=None):
        super().__init__(field, dt, eps, mu, noise_generator)
        self.info['system'] = 'gauss'
        self.fef = GaussianPFCFucntional(eps)

        self._kernel = np.exp(1-field.K2) + field.K2 + 1 - np.e - self.eps
        self._exp_dt_kernel = np.exp(-dt*self._kernel)
        self._mu_dt_half = self.dt * self.mu / 2





from enum import Enum
from pprint import pprint
import warnings
from typing import Tuple, Optional, Union

import numpy as np
from matplotlib import pyplot as plt

from michael960lib.common import deprecated
from torusgrid import fields as fd
from torusgrid.fields import RealField2D
from . import static
from ..core.evolution import PFCMinimizer
from .. import pfc



def generate(na, nb):
    v1 = ((na+nb)*2*np.pi, (na-nb)*2*np.pi/np.sqrt(3))
    v2 = (-v1[1]*np.sqrt(3), v1[0]*np.sqrt(3))

    Ly = np.sqrt(v1[0]**2 + v1[1]**2)
    Lx = np.sqrt(v2[0]**2 + v2[1]**2)

    theta = np.arctan2(v2[1], v2[0])

    return theta, Lx, Ly

def generate_minimal(na: int, nb: int) -> Tuple[float, float, float]:
    # diameter
    D = np.pi * 4 / np.sqrt(3)

    # lattice basis
    a = D*np.array((np.sqrt(3)/2, 1/2))
    b = D*np.array((np.sqrt(3)/2, -1/2))

    Pa = na + 2*nb
    Pb = -2*na - nb

    p0 = np.gcd(Pa, Pb)
    pa = Pa // p0
    pb = Pb // p0
    
    v1 = na*a + nb*b
    v2 = pa*a + pb*b

    theta = np.arctan2(v2[1], v2[0])

    Lx = np.sqrt(v2[0]**2+v2[1]**2)
    Ly = np.sqrt(v1[0]**2+v1[1]**2)

    return theta, Lx, Ly

class AutoDimMode(Enum):
    SCALE = 0
    SNAP = 1

def generate_eps(na: int, nb: int, eps_str: str, 
        NxNy: Union[Tuple[int, int], AutoDimMode]=AutoDimMode.SCALE,
        minimize: Optional[Tuple[float, int]]=None) -> Tuple[float, RealField2D, RealField2D]:
    theta, Lx, Ly = generate_minimal(na, nb)

    sol0 = static.get_relaxed_minimized_coexistent_unit_cell(eps_str, liquid=False)
    liq0 = static.get_relaxed_minimized_coexistent_unit_cell(eps_str, liquid=True)
    mu = static.get_coexistent_mu_final(eps_str)
    eps = float(eps_str)


    Lx0 = sol0.Lx
    Ly0 = sol0.Ly
    Nx0 = sol0.Nx
    Ny0 = sol0.Ny

    _density = np.sqrt(Nx0*Ny0/Lx0/Ly0)


    try:
        if NxNy is AutoDimMode.SCALE:
            Nx = np.rint(Lx0*_density)
            Ny = np.rint(Ly0*_density)

        elif NxNy is AutoDimMode.SNAP:
            Nx = 2 ** int(np.rint(np.log2(Lx*_density)))
            Ny = 2 ** int(np.rint(np.log2(Ly*_density)))
        
        else:
            assert isinstance(NxNy, tuple)
            assert len(NxNy) == 2
            assert type(NxNy[0]) is type(NxNy[1]) is int
            Nx = NxNy[0]
            Ny = NxNy[1]
    except AssertionError:
        raise ValueError('NxNy must be a tuple of 2 integers (Nx, Ny) or an instance of AutoDimMode')


    sol = fd.RealField2D(Lx, Ly, Nx, Ny)
    liq = fd.RealField2D(Lx, Ly, Nx, Ny)

    Xr = np.cos(theta) * sol.X - np.sin(theta) * sol.Y
    Yr = np.sin(theta) * sol.X + np.cos(theta) * sol.Y

    Ir = (Xr / Lx0 * Nx0).astype(int) % Nx0
    Jr = (Yr / Ly0 * Ny0).astype(int) % Ny0
    
    sol_psi = sol0.psi[Ir,Jr] 
    liq_psi = liq0.psi[Ir,Jr] 
    sol.set_psi(sol_psi)
    liq.set_psi(liq_psi)

    dfmt = '[{system}][{label}][t={age:.2f}] f={f:.10f} F={F:.10f} omega={omega:.10f} Omega={Omega:.10f} psibar={psibar:.10f}'
    if minimize is not None:
        model_sol = pfc.PFC(sol)
        model_sol.evolve(minimizer='mu', dt=minimize[0], eps=eps, mu=mu, N_epochs=minimize[1], display_format=dfmt)
        model_liq = pfc.PFC(liq)
        model_liq.evolve(minimizer='mu', dt=minimize[0], eps=eps, mu=mu, N_epochs=minimize[1], display_format=dfmt)
    return theta, sol, liq


# generate a minimized solid profile for mu=0.195 eps=0.1
@deprecated('use generate_eps instead')
def generate_195(na, nb, Nx, Ny):
    theta, Lx, Ly = generate(na, nb)
    unit = np.load('saved_profiles/unit_cell_0.1950_512.npz')

    Lx0 = unit['Lx']
    Ly0 = unit['Ly']
    Nx0 = unit['psi'].shape[0]
    Ny0 = unit['psi'].shape[1]

    x0 = np.linspace(0, Lx0, Nx0)
    y0 = np.linspace(0, Ly0, Ny0)
    X0, Y0 = np.meshgrid(x0, y0, indexing='ij')
    PSI0 = unit['psi']

    x = np.linspace(0, Lx, Nx)
    y = np.linspace(0, Ly, Ny)
    X, Y = np.meshgrid(x, y, indexing='ij')


    Xr = np.cos(theta) * X - np.sin(theta) * Y    
    Yr = np.sin(theta) * X + np.cos(theta) * Y    

    Ir = (Xr / Lx0 * Nx0).astype(int) % Nx0
    Jr = (Yr / Ly0 * Ny0).astype(int) % Ny0
    
    
    PSI = PSI0[Ir,Jr] 
    

    return X, Y, PSI, Lx, Ly, theta, unit

def harmonic(theta, x, y):
    k1 = (np.cos(theta), np.sin(theta))
    k2 = (np.cos(theta+2*np.pi/3), np.sin(theta+2*np.pi/3))
    k3 = (np.cos(theta+4*np.pi/3), np.sin(theta+4*np.pi/3))
    return np.cos(k1[0]*x+k1[1]*y) + np.cos(k2[0]*x+k2[1]*y) + np.cos(k3[0]*x+k3[1]*y)





from typing import List, Dict

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib.widgets import Slider
import matplotlib.gridspec as gridspec

from michael960lib.common import IllegalActionError, with_type
from torusgrid import field_util
from torusgrid import fields as fd
from torusgrid.fields import RealField2D
from .core.base import PFCStateFunction, get_latex
from .core.evolution import PFCMinimizerHistory, import_minimizer_history


_default_colors = ['steelblue', 'darkseagreen', 'palevioletred', 'burlywood']


class PFCHistoryBlock:
    def __init__(self):
        raise NotImplementedError()

    def export_state(self):
        raise NotImplementedError()

    def import_state(self):
        raise NotImplementedError()

    def get_label(self) -> str:
        raise NotImplementedError()

    def has_span(self) -> bool:
        raise NotImplementedError()

    def get_final_field_state(self):
        raise NotImplementedError()

    def get_age_span(self) -> float:   
        raise NotImplementedError()

    def get_state_functions(self)-> Dict[str, List[float]]:
        raise NotImplementedError()

    def get_t(self) -> List[float]:
        raise NotImplementedError()

    def export(self):
        raise NotImplementedError()

class PFCInitialHistoryBlock(PFCHistoryBlock):
    def __init__(self, field: RealField2D, label):
        self.label = label
        self.field = field

    def get_label(self):
        return self.label

    def has_span(self):
        return False

    def get_age_span(self):
        return 0 

    def get_final_field_state(self):
        return self.field.export_state()

    def get_state_functions(self):
        raise IllegalActionError('there is no state functions associated with an edit action')

    def get_t(self):
        raise IllegalActionError('there is no time associated with an edit action')

    @with_type('initial')
    def export(self):
        return {
            'label': self.label,
            'field': self.field.export_state()
        }


class PFCMinimizerHistoryBlock(PFCHistoryBlock):
    def __init__(self, minimizer_history: PFCMinimizerHistory):
        if not minimizer_history.is_committed():
            raise IllegalActionError('cannot initialize history block with uncommitted minimizer history')
        self.minimizer_history = minimizer_history

        self.state_functions = {
            'Lx': [],
            'Ly': [],
            'f': [],
            'F': [],
            'psibar': [],
            'omega': [],
            'Omega': [],
        }

        keys = self.state_functions.keys()

        for sf in self.minimizer_history.get_state_functions():
            for item_name in keys:
                self.state_functions[item_name].append(sf.get_item(item_name))

        self.t = np.array(self.minimizer_history.t)
        self.label = minimizer_history.get_label()

    def get_label(self):
        return self.label

    def get_final_field_state(self):
        return self.minimizer_history.get_final_field_state()

    def has_span(self):
        return True

    def get_age_span(self):
        return self.minimizer_history.age
    
    def get_state_functions(self):
        return self.state_functions
    
    def get_t(self):
        return self.t

    @with_type('minimizer')
    def export(self) -> dict:
        state = {'minimizer_history': self.minimizer_history.export(), 'label': self.label}
        return state


class PFCEditActionHistoryBlock(PFCHistoryBlock):
    def __init__(self):
        pass

    def get_label(self):
        pass

    def has_span(self):
        return False

    def get_age_span(self):
        return 0 

    def get_final_field_state(self):
        pass

    def get_state_functions(self):
        raise IllegalActionError('there is no state functions associated with an edit action')

    def get_t(self):
        raise IllegalActionError('there is no time associated with an edit action')

    @with_type('edit_action')
    def export(self) -> dict:
        raise NotImplementedError()


def import_initial_history_block(state: dict) -> PFCInitialHistoryBlock:
    block = PFCInitialHistoryBlock(fd.import_field(state['field']), state['label'])
    return block


def import_minimizer_history_block(state: dict) -> PFCMinimizerHistoryBlock:
    minimizer_history = import_minimizer_history(state['minimizer_history'])
    block = PFCMinimizerHistoryBlock(minimizer_history)
    return block


def import_edit_action_history_block(state: dict) -> PFCEditActionHistoryBlock:
    raise NotImplementedError()


def import_history_block(state: dict) -> PFCHistoryBlock:
    if state['type'] == 'minimizer':   
        return import_minimizer_history_block(state)
    if state['type'] == 'initial': 
        return import_initial_history_block(state)
    if state['type'] == 'edit_action':   
        return import_edit_action_history_block(state)

    raise ValueError(f'{state["type"]} is not a valid history block type')



class PFCHistory:
    def __init__(self, field: RealField2D):
        self.blocks = []
        self.age_record = []
        self.append(PFCInitialHistoryBlock(field, 'loaded profile'))

    def append(self, block: PFCHistoryBlock):
        if not isinstance(block, PFCHistoryBlock):
            raise ValueError('only PFCHistoryBlock objects can be appended into a PFCHistory')
        self.blocks.append(block)
        max_age = self.get_max_age()

        if block.has_span():
            self.age_record.append(max_age + block.get_age_span())
        else:
            self.age_record.append(max_age)

    def cut_and_insert(self, block: PFCHistoryBlock, index):
        self.blocks = self.blocks[:index]
        self.append(block)
    
    def get_max_age(self):
        max_age = 0
        if len(self.age_record) > 0:
            max_age = self.age_record[-1]
        return max_age 

    def get_blocks(self) -> List[PFCHistoryBlock]:
        return self.blocks.copy()

    def get_block(self, i) -> PFCHistoryBlock:
        return self.blocks[i]

    def get_block_age_lim(self, i):
        if i == 0:
            return 0, 0
        return self.age_record[i-1], self.age_record[i]

    # initial field state of blocks[i]
    def get_field_state(self, i):
        if i == 0:
            return
        return self.blocks[i-1].get_final_field_state()

    def plot(self, *item_names, colors=_default_colors, start=0, end=-1, show=True):
        nrows = len(item_names)
        if nrows == 0:
            raise ValueError('cannot plot history with zero items')

        plt.subplots_adjust(bottom=0.05, top=0.95, left=0.08, right=0.92)
        outer = gridspec.GridSpec(3, 1, height_ratios=[1.5,32,2], hspace=0.1) 
        gs_sliders = gridspec.GridSpecFromSubplotSpec(2, 1, height_ratios=[1,1], subplot_spec=outer[0], hspace=0.2)
        gs_plots = gridspec.GridSpecFromSubplotSpec(nrows, 1, subplot_spec = outer[1], hspace=0.1)
        gs_time = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=outer[2], hspace=0)

        axslider1 = plt.subplot(gs_sliders[0])
        axslider1.xaxis.tick_top()
        axslider2 = plt.subplot(gs_sliders[1])
        axs = [plt.subplot(cell) for cell in gs_plots]
        axT = plt.subplot(gs_time[0])

        time_coords = []
        time_boxes = []

        if end < 0:
            end = len(self.get_blocks()) + end


        color_index = 0
        for j, block in enumerate(self.get_blocks()):
            if j < start or j > end:
                continue
            age_span = block.get_age_span()
            age_i, age_f = self.get_block_age_lim(j)
            rx, ry, rw, rh = age_i, 0, age_f-age_i, 1
            cx, cy = rx+rw/2, ry+rh/2


            rect = Rectangle((rx, ry), rw, rh, facecolor=colors[color_index%len(colors)], edgecolor='none')
            time_coords.append((rx, ry, rw, rh))
            time_boxes.append(rect)

            if block.has_span():
                sfs = block.get_state_functions()
                t = block.get_t() 
                

                axT.add_artist(rect)
                annot_alt = block.get_label().replace(' ', '\n', 1)
                axT.annotate(annot_alt, (cx, cy), color='oldlace', ha='center', va='center', fontsize=10, clip_on=True)


                for i, item_name in enumerate(item_names):
                    ax = axs[i]
                    ax.plot(t+age_i, sfs[item_name], color=colors[color_index%len(colors)], lw=1)
                    ax.set_ylabel(get_latex(item_name))
                    
                    ax.plot([age_i, age_i], [-1e2, 1e2], lw=1, linestyle=':', color='k')

                    if sfs[item_name][0] is None:
                        ax.add_patch(Rectangle((age_i, -1e2), age_f-age_i, 2e2, facecolor='grey', edgecolor='none'))

                color_index += 1

            else:
                pass

        for i, item_name in enumerate(item_names):
            #axs[i].tick_params(axis='x', direction='in', pad=-12)
            #axs[i].tick_params(axis='y', direction='in', pad=-13)
            #axs[i].get_xticklabels()[0].set_ha('left')
            
            ymin = 1e4
            ymax = -1e4
            for j, block in enumerate(self.get_blocks()): 
                if block.has_span():
                    ydata = block.get_state_functions()[item_name]
                    if None in ydata:
                        continue
                    ymin = min(np.min(ydata), ymin)
                    ymax = max(np.max(ydata), ymax)
            
            ydelta = ymax - ymin
            axs[i].set_ylim(ymin - ydelta*0.05, ymax + ydelta*0.05)

            for tick in axs[i].get_yticklabels():
                pass
                #tick.set_ha('left')


        axT.set_yticks([])
        axT.set_xlabel('$t$', fontsize=12)
        
        width = 40

        age_start = self.get_block_age_lim(start)[0]
        age_end = self.get_block_age_lim(end)[1]
        age_range = age_end - age_start
        

        max_width = 150
        zoom_i = 1e-6
        zoom_f = 100-1e-6
        slider2_init = min(max(zoom_i, 100 * (1-max_width/age_range)), zoom_f)

        width = age_range * (100-slider2_init) / 100

        slider1 = Slider(axslider1, '$t$', age_start, age_start+width, valinit=age_start, valfmt='%.3f', color='mediumorchid')
        slider2 = Slider(axslider2, 'zoom', zoom_i, zoom_f, valinit=slider2_init, valfmt=f'%.3f%%', color='plum')

        axslider2.set_xticks([zoom_i, zoom_f], ['-', '+'])

        axslider1.add_artist(axslider1.xaxis)
        axslider2.add_artist(axslider2.xaxis)

        def update1(val):
            width = age_range * (100-slider2.val) / 100
            for i in range(nrows):
                axs[i].set_xlim(slider1.val, slider1.val+width)
            axT.set_xlim(slider1.val, slider1.val + width)

        def update2(val):
            width = age_range * (100-slider2.val) / 100
            for i in range(nrows):
                axs[i].set_xlim(slider1.val, slider1.val+width)
            axT.set_xlim(slider1.val, slider1.val + width)

            slider1_max = max(age_end-width, age_start) 

            slider1_new_val = min(slider1.val, slider1_max)

            slider1.set_val(slider1_new_val)
            slider1.valmax = slider1_max
            axslider1.set_xlim(age_start, slider1_max)

            tmp = slider1_max - age_start
            axslider1.set_xticks([age_start, age_start + tmp/4, age_start + tmp/2, age_start + tmp*3/4, age_start + tmp])

        slider1.on_changed(update1)
        slider2.on_changed(update2)

        update1(age_start)
        update2(slider2_init)

        # tooltip
        annot = axT.annotate("", xy=(0,0), xytext=(0,0), textcoords="offset points",
                bbox=dict(boxstyle="round", fc="w"), arrowprops=dict(arrowstyle="->"), ha='center')
        annot.set_visible(False)

        fig = plt.gcf()
        def update_annot(i, xpos):
            block = self.get_block(i)
            annot.set_text(block.get_label()) 
            annot.xy = (xpos, time_coords[i][1] + time_coords[i][3])
            annot.get_bbox_patch().set_alpha(0.9)

        def on_hover(event):
            vis = annot.get_visible()
            if event.inaxes == axT:
                for i, time_box in enumerate(time_boxes):
                    cont = time_box.contains(event)
                    if cont[0]:
                        xpos = event.xdata
                        update_annot(i, xpos)
                        annot.set_visible(True)
                        fig.canvas.draw_idle()
                        return
                if vis:
                    annot.set_visible(False)
                    fig.canvas.draw_idle()
            else:
                if vis:
                    annot.set_visible(False)
                    fig.canvas.draw_idle()



        fig.canvas.mpl_connect('motion_notify_event', on_hover)

        mng = plt.get_current_fig_manager()
        mng.resize(*mng.window.maxsize())
        widgets = [slider1, slider2]

        if show:
            plt.show()
        else:
            return axT, axs, axslider1, axslider2, widgets 

    def export(self):
        return {
            'blocks': [block.export() for block in self.blocks],
            'age_record': self.age_record,
        }


def import_history(state: dict) -> PFCHistory:
    initial_field = import_initial_history_block(state['blocks'][0]).field
    pfc_history = PFCHistory(initial_field)

    pfc_history.blocks = [import_history_block(block_state) for block_state in state['blocks']]
    pfc_history.age_record = state['age_record']

    return pfc_history

